# std
import os
import yaml

# internal
from laz.utils.logging import get_logger
from laz.utils import log
from laz.cli.parser import parser


subparsers = parser.add_subparsers(required=True)

parser_init = subparsers.add_parser('init')

def init():
    args = parser.parse_args()
    get_logger(verbosity=args.verbose)

    if os.path.isfile('laz.yml'):
        log.error('laz.yml exists')
        exit(1)

    data = yaml.safe_dump({
        'env': {
            'example': 'value',
        },
        'targets': {
            'dev': None,
        },
        'actions': {
            'example': 'echo $example',
            'hello': 'echo {{ target.name }}',
        },
    }, sort_keys=False).replace('null', '')

    with open('laz.yml', 'w') as fh:
        fh.write(data)



# std
from typing import List

# internal
from laz.model.tree import Node
from laz.model.path import Path
from laz.model.resolver import Resolver
from laz.model.target import Target
from laz.model.act import Act


class Runner:

    def __init__(self, root_node: Node, args: List[str]):
        self.root_node = root_node
        self.path = Path(args[0])
        self.args = args[1:]

    def resolve(self) -> List[Target]:
        resolver = Resolver(self.root_node, self.path)
        return resolver.resolve()

    def run(self):
        targets = self.resolve()
        for target in targets:
            act = Act(target, self.args)
            act.act()



class LazError(Exception):
    pass


class LazValueError(LazError, ValueError):
    pass


class LazTypeError(LazError, TypeError):
    pass


class LazRuntimeError(LazError, RuntimeError):
    pass



# std
import sys
from typing import Optional as Opt

# internal
from laz.utils import log


def main():
    log.debug(sys.argv[1:])
    subcommand = _subcommand()
    if len(sys.argv) == 2 and (sys.argv[1] == 'version' or sys.argv[1] == '--version'):
        from laz.cli.subcommands.version import version
        version()
    elif subcommand == 'init':
        from laz.cli.subcommands.init import init
        init()
    elif subcommand == 'tree':
        from laz.cli.subcommands.tree import tree
        tree()
    else:
        from laz.cli.root import root
        root()


def _subcommand() -> Opt[str]:
    for i, s in enumerate(sys.argv):
        if i > 0 and not s.startswith('-'):
            return sys.argv[i]


if __name__ == '__main__':
    cmd = 'laz version'
    sys.argv = cmd.split(' ')
    main()



# std
from contextlib import contextmanager
import os


@contextmanager
def in_dir(dirpath: str):
    previous_dirpath = os.getcwd()
    try:
        os.chdir(dirpath)
        yield
    finally:
        os.chdir(previous_dirpath)



# std
from __future__ import annotations
from abc import abstractmethod
import os
import subprocess
from typing import List, Union

# internal
from laz.utils.errors import LazValueError
from laz.utils import log
from laz.utils.types import Data, DictData, ListData
from laz.model.target import Target


class Action:
    action_types = []

    def __init__(self, target: Target, run_data: Data):
        self.target: Target = target
        self.run_data: Data = run_data

    def __init_subclass__(cls):
        if cls.__name__ != 'ShellAction':
            cls.action_types.append(cls)

    @abstractmethod
    def run(self):
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def is_handler(cls, target: Target, run_data: Data) -> bool:
        raise NotImplementedError

    @classmethod
    def new(cls, target: Target, run_data: Data) -> Action:
        for action_type in cls.action_types:
            if action_type.is_handler(target, run_data):
                return action_type.new(target, run_data)
        if ShellAction.is_handler(target, run_data):
            return ShellAction.new(target, run_data)
        cls.raise_value_error(target, run_data)

    @classmethod
    def raise_value_error(cls, target: Target, run_data: Data):
        log.error(f'target -> {target}')
        log.error(f'run data -> {run_data}')
        raise LazValueError(f'Given values could not be handled as a {cls.__name__}')


class ConfiguredAction(Action):

    def run(self):
        raise NotImplementedError

    @classmethod
    def is_handler(cls, target: Target, run_data: Data) -> bool:
        configured_actions: DictData = target.data.get('actions', {})
        if isinstance(run_data, str) and run_data in configured_actions:
            return True
        else:
            return False

    @classmethod
    def new(cls, target: Target, run_data: Data) -> Action:
        configured_actions: DictData = target.data.get('actions', {})
        if isinstance(run_data, str) and run_data in configured_actions:
            return Action.new(target, configured_actions[run_data])
        else:
            cls.raise_value_error(target, run_data)


class MultipleActions(Action):

    def __init__(self, target: Target, run_data: ListData):
        super().__init__(target, run_data)
        self.actions = [Action.new(target, d) for d in self.run_data]

    def run(self):
        results = []
        for action in self.actions:
            results.append(action.run())
        return results

    @classmethod
    def is_handler(cls, target: Target, run_data: Data) -> bool:
        return isinstance(run_data, list)

    @classmethod
    def new(cls, target: Target, run_data: ListData) -> Action:
        return MultipleActions(target, run_data)


class ConditionalAction(Action):

    def run(self):
        result = Action.new(self.target, self.run_data['condition']).run()
        if bool(result):
            if 'if' in self.run_data:
                return Action.new(self.target, self.run_data['if']).run()
        else:
            if 'else' in self.run_data:
                return Action.new(self.target, self.run_data['else']).run()

    @classmethod
    def is_handler(cls, target: Target, run_data: Data) -> bool:
        return isinstance(run_data, dict) and 'condition' in run_data

    @classmethod
    def new(cls, target: Target, run_data: DictData) -> Action:
        return cls(target, run_data)


class SwitchAction(Action):

    def run(self):
        result = Action.new(self.target, self.run_data['switch']).run()
        if result in self.run_data:
            return Action.new(self.target, self.run_data[result]).run()
        elif 'default' in self.run_data:
            return Action.new(self.target, self.run_data['default']).run()
        else:
            return

    @classmethod
    def is_handler(cls, target: Target, run_data: Data) -> bool:
        return isinstance(run_data, dict) and 'switch' in run_data

    @classmethod
    def new(cls, target: Target, run_data: DictData) -> Action:
        return cls(target, run_data)


class ShellAction(Action):
    run_data: DictData

    def run(self):
        env = {
            **os.environ,
            **self.target.data.get('env', {}),
        }
        if self.run_data.get('capture'):
            type_str = self.run_data.get('type', 'str')
            result = subprocess.run(self.args(), env=env, capture_output=True).stdout.decode()
            if self.run_data.get('strip', True):
                result = result.strip()
            result = eval(f'{type_str}({result})')
            return result
        else:
            return subprocess.run(self.args(), env=env).returncode == 0

    def args(self) -> List[str]:
        return [self._shell(), '-c', self.run_data['command']]

    def _shell(self) -> str:
        return os.environ.get('SHELL', 'bash')

    @classmethod
    def is_handler(cls, target: Target, run_data: Data) -> bool:
        if isinstance(run_data, str):
            return True
        elif isinstance(run_data, dict) and 'command' in run_data:
            return True
        else:
            cls.raise_value_error(target, run_data)

    @classmethod
    def new(cls, target: Target, run_data: Union[str, DictData]) -> Action:
        if isinstance(run_data, str):
            run_data = {'command': run_data}
        return cls(target, run_data)


class PythonAction(Action):

    def run(self):
        return eval(self.run_data['python'])

    @classmethod
    def is_handler(cls, target: Target, run_data: Data) -> bool:
        return isinstance(run_data, dict) and 'python' in run_data

    @classmethod
    def new(cls, target: Target, run_data: DictData) -> Action:
        return cls(target, run_data)



# std
from copy import deepcopy
import itertools

# internal
from laz.utils.types import AtomicData, ListData, DictData, Data
from laz.utils.errors import LazTypeError


def merge(left: Data, right: Data) -> Data:
    if type(left) != type(right):
        raise LazTypeError(f'merge inputs must be same type. Got {type(left)} and {type(right)}')
    return _merge(deepcopy(left), deepcopy(right))


def _merge(left: Data, right: Data) -> Data:
    if isinstance(left, dict) and isinstance(right, dict):
        return _merge_dicts(left, right)
    elif isinstance(left, list) and isinstance(right, list):
        return _merge_lists(left, right)
    elif isinstance(left, (type(None), bool, int, str)) and isinstance(right, (type(None), bool, int, str)):
        return _merge_atomics(left, right)
    elif left is None:
        return right
    elif right is None:
        return left
    else:
        raise LazTypeError(f'merge inputs must be same type. Got {type(left)} and {type(right)}')


def _merge_dicts(left: DictData, right: DictData) -> DictData:
    for key in right.keys():
        l = left.get(key)
        r = right.get(key)
        left[key] = _merge(l, r)
    return left


def _merge_lists(left: ListData, right: ListData) -> ListData:
    new_list = [None] * max(len(left), len(right))
    for i, (l, r) in enumerate(itertools.zip_longest(left, right)):
        new_list[i] = _merge(l, r)
    return new_list


def _merge_atomics(left: AtomicData, right: AtomicData) -> AtomicData:
    if left is None and right is not None:
        return right
    elif left is not None and right is None:
        return left
    return right



# internal
from laz.utils.logging import get_logger
from laz.cli.parser import parser
from laz.utils.load import load
from laz.model.tree import Node
from laz.utils.iterator import Iterator


subparsers = parser.add_subparsers(required=True)

parser_init = subparsers.add_parser('tree')


def tree():
    args = parser.parse_args()
    get_logger(verbosity=args.verbose)
    root = load()
    print(root.configuration.name)
    for child in Iterator(root.children):
        _tree(child.value, '', child.is_last)


def _tree(node: Node, prefix: str = '', last: bool = False):
    symbol = '└─' if last else '├─'
    print(f'{prefix}{symbol} {node.configuration.name}')

    for child in Iterator(node.children):
        _tree(child.value, prefix + '│  ', child.is_last)
